/*
 * DALLOC.C - debugging versions of potentially mis-used standard functions.
 *
 * Limitations:
 *
 * 1. depends on MS-C dynamic allocation structure, so may not be portable
 *    to other compilers.
 * 2. if dfree() is called for storage not allocated by dmalloc(), dcalloc()
 *    or drealloc(), an error may result.
 * 3. contains no debugging analogs for MS-C halloc()/hfree() functions;
 *    d_halloc()/d_hfree() calls devolve into halloc()/hfree() calls in
 *    both debug and production modes.
 *
 * $Header:   C:/fe/shared/vcs/dalloc.c_v   1.3   15 May 1990 17:22:26   rji  $
 *
 * $Log:   C:/fe/shared/vcs/dalloc.c_v  $
 * 
 *    Rev 1.3   15 May 1990 17:22:26   rji
 * Implementation of detection of freeing of freed storage altered for
 * greater flexibility in combining debugging and production modules
 * for testing.  Documented limitations.  Added and improved other comments.
 * All functions now sport ANSI function headers.  Fixed hole in drealloc() 
 * logic (it would have failed if new size < old size).  Added new test
 * driver.  Changed #include file specifiers from <...> to "..." for 
 * non-standard headers.
 * 
 *    Rev 1.2   08 May 1990 23:41:06   rji
 * Added protection against freeing already-freed storage.  (KGW)
 * 
 *    Rev 1.1   14 Aug 1989 11:07:48   rji
 * Public functions given shorter names for portability.
 * 
 *    Rev 1.0   07 Dec 1988 23:09:30   rji
 * Initial revision.
 * 
*/

/*=========================================================================*/
/*=                                                                       =*/
/*=      I N C L U D E S,  D E F I N E S,  D E C L A R A T I O N S        =*/
/*=                                                                       =*/
/*=========================================================================*/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         DALLOC.C Includes

  $PATHS$
       Functions\C
       Shared\Includes
       Shared\Utils
       Modules\$./FL$
   $3$

   See Also:
        $/SEE(DEBUG.H Macros)$
        $/SEE(COMMON.H Macros)$
        $/SEE(DALLOC.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <dos.h>

#include "debug.h"
#include "common.h"
#include "dalloc.h"
/*                                                                   $end$ */

/*-------------------------------------------------------------------------*/
/* MANIFEST CONSTANTS                                                      */
/*-------------------------------------------------------------------------*/

/* memory block status byte values */
#define  BLOCK_ALLOCATED   '\xAA'
#define  BLOCK_FREED       '\xFF'

/*-------------------------------------------------------------------------*/
/* PRIVATE FUNCTIONS                                                       */
/*-------------------------------------------------------------------------*/

static void log_activity( char *, ... );

/*-------------------------------------------------------------------------*/
/* PRIVATE DATA                                                            */
/*-------------------------------------------------------------------------*/

static unsigned int     key = 0;       /* incremented to form a unique id for
                                       ** each memory block allocated --
                                       ** if more than 64K-1 blocks are 
                                       ** allocated, id's will repeat
                                       */

static unsigned long    ncalls   = 0L; /* # successful allocation calls    */

static FILE    *dlog = NULL;           /* for debugging log file           */

static char    drive[_MAX_DRIVE],      /* pathname pieces                  */
               dir[_MAX_DIR],
               fname[_MAX_FNAME],
               ext[_MAX_EXT];

/* the following variables are reserved for future enhancements */
#if 0
static unsigned long    totfree,       /* total free memory at 1st call    */
                        freenow,       /* free memory after nth call       */
                        totused,       /* totfree - freenow                */
                        prevfree,      /* free after previous call         */
                        totasked = 0L, /* sum of malloc - free requests    */
                        totalloc = 0L; /* total actually used              */
#endif


/*=========================================================================*/
/*=                                                                       =*/
/*=                F U N C T I O N   D E F I N I T I O N S                =*/
/*=                                                                       =*/
/*=========================================================================*/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dmalloc()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DMALLOC                                                                 */
/*-------------------------------------------------------------------------*/

void  *dmalloc(
   size_t   nbytes,     /* in: number of bytes to be allocated */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   static char stub[] = "ma";             /* id for debugging messages     */
   char       *memptr;                    /* return from malloc            */
   size_t      len;

   /* break up source path name into manageable pieces */
   _splitpath( pname, drive, dir, fname, ext );

   /* allocate requested memory + storage for key and status byte */
   if ( nbytes > (((size_t) -1) - sizeof(unsigned) - 1) /* assumes 2's comp */
       || (memptr = malloc( nbytes + sizeof(unsigned) + 1)) == NULL )
   {
      fprintf( stderr, "dmalloc: cannot allocate %u bytes requested "
               "in file %s%s, line %d\n", nbytes, fname, ext, linenum );
      exit( 1 );
   }

   ++ncalls;   /* bump successful call counter */

   /* extract length of storage actually allocated by malloc() */
   len = *(size_t *) (memptr - sizeof(size_t));

   /* mark allocated storage with unique key and status */
   *(unsigned *) (memptr + len - sizeof(unsigned)) = ++key;
   *(memptr + len - sizeof(unsigned) - 1) = BLOCK_ALLOCATED;

   /* log activity on debug file */
   log_activity( "%s %8s%-4s %-5d %p %-5u %-5u\n", stub, fname, ext,
     linenum, fptr(memptr), *(unsigned *)(memptr+len-sizeof(unsigned)), len );

   /* return to caller */
   return ( memptr );                     /* give user ptr to his request  */

} /* dmalloc() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dcalloc()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DCALLOC                                                                 */
/*-------------------------------------------------------------------------*/

void  *dcalloc(
   size_t   neles,      /* in: number of elements to allocate  */
   size_t   nbytes,     /* in: size of each element            */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   char    *memptr;                       /* return from calloc            */
   size_t   len   = neles * nbytes;       /* amount of storage requested   */

   /* allocate requested memory & initialize it to zeroes */
   memptr = (char *) dmalloc( len, pname, linenum );
   memset( memptr, '\0', len );

   /* return to caller */
   return ( memptr );                     /* give user ptr to his request  */

} /* dcalloc() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         drealloc()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DREALLOC                                                                */
/*-------------------------------------------------------------------------*/

void  *drealloc(
   char    *oldbuf,     /* in: pointer to buffer to reallocate */
   size_t   nbytes,     /* in: size of buffer when reallocated */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   char    *memptr;                       /* return from realloc           */
   size_t   oldlen;                       /* length of oldbuf -- includes key
                                          ** and status byte if oldbuf was 
                                          ** alloc'd by one of the d*() funcs
                                          */

   /* extract length of buffer to be re-allocated */
   oldlen = *(size_t *) (oldbuf - sizeof(size_t));

   /* reallocate new buffer as requested */
   memptr = (char *) dmalloc( nbytes, pname, linenum );

   /* copy data to new buffer (which may be smaller) */
   dmemcpy( memptr, oldbuf, min( oldlen, nbytes ), pname, linenum );

   /* free old buffer */
   dfree( oldbuf, pname, linenum  );

   /* return to caller */
   return ( memptr );                     /* give user ptr to his request  */

} /* drealloc() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dfree()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DFREE                                                                   */
/*-------------------------------------------------------------------------*/

void  dfree(
   void    *ptr,        /* in: pointer to buffer to be freed   */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   char          *memptr;                    /* local copy of ptr          */
   char          *statptr;                   /* pointer to status byte     */
   size_t         len;
   unsigned       key;
   static char    stub[] = "f ";             /* id for debugging messages  */

   /* check for premature call */
   if ( ncalls < 1 )
   {
      fprintf( stderr, "dfree: Invoked without prior dmalloc call\n" );
      exit( 1 );
   }

   /* validate input argument */
   if ( (memptr = ptr) == NULL )
   {
      fprintf( stderr, "dfree: Attempt to free NULL pointer\n" );
      exit( 1 );
   }

   /* break up source path name into manageable pieces */
   _splitpath( pname, drive, dir, fname, ext );

   /* extract length & key from dynamically allocated buffer */
   len = *(size_t *) (memptr - sizeof(size_t));
   key = *(unsigned *) (memptr + len - sizeof(unsigned));

   /* check status byte to see if OK -- */
   statptr = memptr + len - sizeof(unsigned) - 1;
   if ( *statptr != BLOCK_ALLOCATED )
   {
      fprintf( stderr,
              "dfree: Attempt to free freed storage, bad pointer (%p),\n"
              "storage not allocated by dmalloc, or other error\n",
               fptr(memptr) );
      /* write error prefix for next activity log record */
      log_activity( "Error!" );
   }
   else
   {
      /*
      ** we try not to overwrite storage if it wasn't allocated by dmalloc(),
      ** though it is possible that the BLOCK_ALLOCATED flag is present in
      ** storage not allocated by dmalloc().
      */
      *statptr = BLOCK_FREED; /* change status to freed */
   }

   /* log activity on debug file -- do before free() in case of trouble */
   log_activity( "%s %8s%-4s %-5d %p %-5u %-5u\n",
                 stub, fname, ext, linenum, fptr(memptr), key, len );

   free( memptr ); /* free memory */

   /* return to caller */
   return;

} /* dfree() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dstrcpy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DSTRCPY                                                                 */
/*-------------------------------------------------------------------------*/

char  *dstrcpy(
   char *dest,       /* in: pointer to output string        */
   char *src,        /* in: pointer to input string         */
   char *pname,      /* in: path name of calling module     */
   int   linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   static char    stub[] = "s ";             /* id for debugging messages  */

   /* log activity on debug file */
   _splitpath( pname, drive, dir, fname, ext );
   log_activity( "%s %8s%-4s %-5d %p %p\n", stub, fname, ext, linenum,
                                          fptr(dest), fptr(src) );

   return ( strcpy( dest, src ) );

} /* dstrcpy() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dstrncpy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DSTRNCPY                                                                */
/*-------------------------------------------------------------------------*/

char  *dstrncpy(
   char    *dest,       /* in: pointer to output string        */
   char    *src,        /* in: pointer to input string         */
   size_t   size,       /* in: # of characters to copy         */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   static char    stub[]="sn";               /* id for debugging messages  */

   /* log activity on debug file */
   _splitpath( pname, drive, dir, fname, ext );
   log_activity( "%s %8s%-4s %-5d %p %p %-5u\n",
                  stub, fname, ext, linenum, fptr(dest), fptr(src), size );

   return ( strncpy( dest, src, size ) );

} /* dstrncpy() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         dmemcpy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DMEMCPY                                                                 */
/*-------------------------------------------------------------------------*/

void  *dmemcpy(
   void    *dest,       /* in: pointer to output area          */
   void    *src,        /* in: pointer to input area           */
   size_t   size,       /* in: # of characters to copy         */
   char    *pname,      /* in: path name of calling module     */
   int      linenum     /* in: line number of call w/i module  */
)
/*                                                                   $end$ */
{
   static char    stub[] = "m ";             /* id for debugging messages  */

   /* log activity on debug file */
   _splitpath( pname, drive, dir, fname, ext );
   log_activity( "%s %8s%-4s %-5d %p %p %-5u\n",
                  stub, fname, ext, linenum, fptr(dest), fptr(src), size );

   return ( memcpy( dest, src, size ) );

} /* dmemcpy() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         log_activity()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* LOG_ACTIVITY                                                            */
/*-------------------------------------------------------------------------*/

static void log_activity( char *fmt, ... )
/*                                                                   $end$ */
{
   va_list lp;

   /*
   ** Activity records are only written to the debugging log file
   ** if debugging is activated via set_debug()
   */

   D(
      if ( dlog == NULL )
         if ( (dlog = fopen( "debug.dta", "w" )) == NULL )
         {
            fprintf( stderr, "log_activity: cannot open \"debug.dta\"\n" );
            exit( 1 );
         }
      va_start( lp, fmt );
      vfprintf( dlog, fmt, lp );
      fflush( dlog );
      va_end( lp );
   )

} /* log_activity() */


#if defined (DRIVER)
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         DALLOC.C main() Test Driver

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*=========================================================================*/
/*=                                                                       =*/
/*=                      T E S T   D R I V E R                            =*/
/*=                                                                       =*/
/*=========================================================================*/

void  main( void );

void  main()
/*                                                                   $end$ */
{
   enum optypes
   {
      QUIT, MALLOC, CALLOC, REALLOC, FREE, STRCPY, STRNCPY, MEMCPY
   } op = -1;
   size_t   nelements,
            requested_size,
            allocated_size;
   char    *p = NULL,
            s[1001],
            t[100];

   set_debug( 1 );   /* insure debugging log records are written */

   for (;;)
   {
      nelements = requested_size = 0;

      printf( "\nEnter operation # [and operands, as indicated] where\n\n"
              "%d = QUIT\n"
              "%d = ALLOCATE MEMORY <size>\n"
              "%d = ALLOCATE & CLEAR MEMORY <#elements> <element size>\n"
              "%d = RE-ALLOCATE MEMORY <new size>\n"
              "%d = FREE MEMORY\n"
              "%d = COPY STRING TO MEMORY <string>\n"
              "%d = COPY CHARS FROM STRING TO MEMORY <string> <#chars>\n"
              "%d = COPY BYTES TO MEMORY <bytes_to_copy> <#bytes>\n"
              "\n==> ",
              QUIT, MALLOC, CALLOC, REALLOC, FREE, STRCPY, STRNCPY, MEMCPY );
      fflush( stdout );

      scanf( "%d", &op );  getchar();
      if ( op == QUIT )
         break;

      switch ( op )
      {
         case MALLOC:
            scanf( "%u", &requested_size );  gets( t );
            p = d_malloc( requested_size );
            break;
         case CALLOC:
            scanf( "%u %u", &nelements, &requested_size );  gets( t );
            p = d_calloc( nelements, requested_size );
            break;
         case REALLOC:
            scanf( "%u", &requested_size );  gets( t );
            p = d_realloc( p, requested_size );
            break;
         case FREE:
            /* two consecutive FREE's should produce log Error! record */
            d_free( p );
            break;
         case STRCPY:
            scanf( "%1000s", s );  gets( t );
            d_strcpy( p, s );
            break;
         case STRNCPY:
            scanf( "%1000s %u", s, &requested_size );  gets( t );
            d_strncpy( p, s, requested_size );
            break;
         case MEMCPY:
            scanf( "%1000s %u", s, &requested_size );  gets( t );
            d_memcpy( p, s, requested_size );
            break;
         default:
            fprintf( stderr, "\nInvalid selection --try again!\n" );
      } /* switch */

      /* dump allocated storage -- note MS-C free() adds 1 to size prefix */
      if ( p != NULL && op >= QUIT && op <= MEMCPY )
      {
         allocated_size = *(size_t *)(p-sizeof(size_t));
         hexdump( p, ((op != FREE) ? allocated_size : allocated_size -1) );
      }

   } /* for */

   exit( 0 );

} /* main() */

#endif /* if defined(DRIVER) */
